//
// Fl_Table -- A table widget
//
// Copyright 2002 by Greg Ercolano.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems to "erco@seriss.com".
//
// TODO:
//    o Auto scroll during dragged selection
//    o Keyboard navigation (up/down/left/right arrow)
//

#ifndef _FL_TABLE_H
#define _FL_TABLE_H

#include <FL/Fl.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Scrollbar.H>
#include <sys/types.h>
#include <vector>
#include <iostream.h>		// I YAM HERE

class Fl_Table : public Fl_Group
{
public:
    enum TableContext
    {
	CONTEXT_NONE		= 0,
	CONTEXT_NEWPAGE		= 1,	// before a page is redrawn
	CONTEXT_ROW_HEADER	= 2,	// in the row header
	CONTEXT_COL_HEADER	= 3,	// in the col header
	CONTEXT_CELL		= 4,	// in one of the cells
	CONTEXT_TABLE		= 5,	// in the table
	CONTEXT_RC_RESIZE	= 6	// column or row being resized
    };

private:
    int _rows, _cols,	 	// total rows/cols
	_row_header_w,		// width of row header
	_col_header_h;		// height of column header

    char _row_header,		// row header enabled?
         _col_header,		// col header enabled?
	 _row_resize,		// row resizing enabled?
	 _col_resize;		// col resizing enabled?

    Fl_Color _row_header_color,
             _col_header_color;

    std::vector<int> 
         _colwidths,		// column widths in pixels (_colw must be -1)
    	 _rowheights;		// row heights in pixels (_rowh must be -1)

    Fl_Cursor last_cursor;	// last mouse cursor before changed to 'resize' cursor
    Fl_When _when;		// when to do callback (FL_WHEN_RELEASE, etc)

    void fix_scrollbar_order();	// force scrollbars to be last children
    				// so they receive events before others

protected:
    enum ResizeFlag
    {
        RESIZE_NONE      = 0,
        RESIZE_COL_LEFT  = 1,
        RESIZE_COL_RIGHT = 2,
        RESIZE_ROW_ABOVE = 3,
        RESIZE_ROW_BELOW = 4
    };

    int _tablew, _tableh;			// table's virtual size (in pixels)
    int toprow, botrow, 			// four corners of viewable table
	leftcol, rightcol;
    Fl_Box       *tablebox;			// invisible box for table area
    Fl_Scrollbar *vscrollbar, 			// vertical scrollbar
                 *hscrollbar;			// horizontal scrollbar

    // Fltk
    int handle(int e);				// fltk handle() override

    // Class maintenance
    void table_window_size(int &X, int &Y, 	// return size of table window
    			   int &W, int &H);
    void table_resized();			// table resized; recalc
    void table_scrolled();			// table scrolled; recalc
    void get_bounds(TableContext context,	// return x/y/w/h bounds for context
                    int &X, int &Y, int &W, int &H);
    void change_cursor(Fl_Cursor newcursor);	// change mouse cursor to some other shape
    Fl_Table::TableContext cursor2rowcol(	// find r/c given current x/y event
    		    int &R, int &C, ResizeFlag &resizeflag);
    int find_cell(TableContext context,		// find cell's x/y/w/h given r/c
    		 int R, int C, int &X, int &Y, int &W, int &H);
    int row_col_clamp(int &R, int &C);		// clamp r/c to known universe

    // Called to draw cells
    virtual void draw_cell(TableContext context, int R=0, int C=0, 
    			   int X=0, int Y=0, int W=0, int H=0)
	{ }					// overridden by deriving class

    // Event callback
    virtual void callback(TableContext context, int R, int C)
	{ }	 				// overridden by deriving class

    long row_scroll_position(int row);		// find scroll position of row (in pixels)
    long col_scroll_position(int col);		// find scroll position of col (in pixels)

    int is_fltk_container() 			// does table contain fltk widgets?
        { return( Fl_Group::children() > 3 ); }	// (ie. more than box and 2 scrollbars?)

    static void scroll_cb(Fl_Widget*,void*);	// h/v scrollbar callback

public:
    Fl_Table(int X, int Y, int W, int H, const char *l=0);
    ~Fl_Table();

    int x() { return(Fl_Group::x() + 1); }	// +1: widget border compensation
    int y() { return(Fl_Group::y() + 1); }	// +1: widget border compensation
    int w() { return(Fl_Group::w() - 2); }	// -2: widget border compensation
    int h() { return(Fl_Group::h() - 2); }	// -2: widget border compensation

    void rows(int val);				// set/get number of rows
    inline int rows() 
        { return(_rows); }

    void cols(int val);				// set/get number of columns
    inline int cols()
        { return(_cols); }

    inline int row_resize()
        { return(_row_resize); }
    void row_resize(int flag)			// enable row resizing
        { _row_resize = flag; }

    inline int col_resize()
        { return(_col_resize); }
    void col_resize(int flag)			// enable col resizing
        { _col_resize = flag; }

    inline int row_header()			// set/get row header enable flag
        { return(_row_header); }
    void row_header(int flag)
        { _row_header = flag; table_resized(); redraw(); }

    inline int col_header()			// set/get col header enable flag
        { return(_col_header); }
    void col_header(int flag)
        { _col_header = flag; table_resized(); redraw(); }

    inline void col_header_height(int height)	// set/get col header height
        { _col_header_h = height; table_resized(); redraw(); }
    inline int col_header_height()
        { return(_col_header_h); }

    inline void row_header_width(int width)	// set/get row header width
        { _row_header_w = width; table_resized(); redraw(); }
    inline int row_header_width()
        { return(_row_header_w); }

    inline void row_header_color(Fl_Color val)	// set/get row header color
        { _row_header_color = val; redraw(); }
    inline Fl_Color row_header_color()
        { return(_row_header_color); }

    inline void col_header_color(Fl_Color val)	// set/get col header color
        { _col_header_color = val; redraw(); }
    inline Fl_Color col_header_color()
        { return(_col_header_color); }

    void row_height(int row, int height);	// set/get row height
    inline int row_height(int row)
	{ return((row<0 || row>=(int)_rowheights.size()) ? 0 : _rowheights[row]); }

    void col_width(int col, int width);		// set/get a column's width
    inline int col_width(int col)
	{ return((col<0 || col>=(int)_colwidths.size()) ? 0 : _colwidths[col]); }

    void row_height_all(int height)		// set all row/col heights
        { for ( int r=0; r<rows(); r++ ) row_height(r, height); }
    void col_width_all(int width)
        { for ( int c=0; c<cols(); c++ ) col_width(c, width); }

    void row_position(int row);			// set/get table's current scroll position
    void col_position(int col);
    int row_position()				// current row position
	{ return(toprow); }
    int col_position()				// current col position
	{ return(leftcol); }

    inline void top_row(int row)		// set/get top row (deprecated)
        { row_position(row); }
    inline int top_row()
        { return(row_position()); }

    // Fltk overrides / callbacks
    void when(Fl_When val)			// set/get 'when' table receives events
	{ _when = val; }
    inline Fl_When when()
        { return(_when); }

    void redraw()				// redraw everything
    {
        vscrollbar->redraw();
        hscrollbar->redraw();
        Fl_Group::redraw();
    }
    void resize(int X, int Y, int W, int H);	// fltk resize() override
    void draw(void);				// fltk draw() override
    void printself();				// debugging
};

#endif /*_FL_TABLE_H*/
